#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <stdio.h>
#include "path.h"
#include "bfile.h"
#include "path.h"
#include "main.h"

void loprintf(const char *fmt, ...);  

static void strlist_init(strlist_t *K)
{
  K->prev= K->next= K;
  K->name= NULL;
}

void strnode_remove(strlist_t *K)
{
  K->prev->next= K->next;
  K->next->prev= K->prev;
}

static void strnode_insert_before(strlist_t *anchor, strlist_t *N)
{
  N->prev= anchor->prev;
  N->next= anchor;
  N->prev->next= N;
  anchor->prev= N;
}

static strlist_t *strnode_new(char *data)
{
  strlist_t *K;
  K= calloc(1,sizeof(*K));
  K->name= strdup(data);
  return K;
}


int bfile_has_perm(bfile_t *bf, char *name)
{
  strlist_t *Z;
  for(Z=bf->perms.next;Z!=&bf->perms;Z=Z->next)
    if (!strcmp(Z->name,name)) return 1;
  return 0;
}

void bfile_add_perm(bfile_t *bf, char *name)
{
  strlist_t *K,*Z;
  K= strnode_new(name);
  for(Z=bf->perms.next;Z!=&bf->perms;Z=Z->next)
  {
    if (strcmp(Z->name,K->name)<0) break;
  }
  strnode_insert_before(Z,K);
}

void bfile_remove_perm(bfile_t *bf, char *name)
{
  strlist_t *Z;
  for(Z=bf->perms.next;Z!=&bf->perms;Z=Z->next)
  {
    if (!strcmp(Z->name,name)) break;
  }
  if (Z==&bf->perms) return ;
  strnode_remove(Z);
  free(Z->name);
  free(Z);
}


c_source_t* bfile_add_c_source(bfile_t *bf,char *fn,int n)
{
  c_source_t *S;
  S= calloc(1,sizeof(*S));
  S->name= strdup(fn);
  S->dhead.next= S->dhead.prev= &S->dhead;
  S->dhead.name= NULL;
  if (n<0)
  {
    int max= 0;
    c_source_t *K;
    for(K=bf->c_sources.next;K!=&bf->c_sources;K=K->next)
      if (K->output_no>max) max= K->output_no;
    n= max+1;
  }
  S->output_no= n;
  S->next= &bf->c_sources;
  S->prev= bf->c_sources.prev;
  bf->c_sources.prev= S;
  S->prev->next= S;
  return S;
}

c_dep_t* bfile_add_c_dependency(bfile_t *bf, c_source_t *cs,char *fname)
{
  c_dep_t *cd;
  cd= malloc(sizeof(*cd));
  cd->name= strdup(fname);
  cd->prev= cs->dhead.prev;
  cd->next= &cs->dhead;
  cs->dhead.prev= cd;
  cd->prev->next= cd;
  return cd;
}

static char *template_fn()
{
  char *home;
  char buf[2000];
  home= getenv("HOME");
  if (home)
    sprintf(buf,"%s/.config/abu/template", home);
  else
    strcpy(buf,"/etc/abu/template");
  return strdup(buf);
}


strlist_t *bfile_add_c_include(bfile_t *bf, char *name)
{
  strlist_t *K;
  K= strnode_new(name);
  strnode_insert_before(&bf->c_include_paths, K);
  return K;
}

strlist_t *bfile_add_c_library(bfile_t *bf, char *name)
{
  strlist_t *K;
  K= strnode_new(name);
  strnode_insert_before(&bf->c_library_paths, K);
  return K;
}


static void prt_cmd(FILE *f, char *cmd, char *val)
{
  int i;
  for(i=0;val[i];i++) if (val[i]=='\n') val[i]=' ';
  fprintf(f, "%s %s\n", cmd, val);
}

static int get_cmd(FILE *f, char *cmd, char *val)
{
  char buf[4000];
  int i,j;
  if (!fgets(buf,sizeof(buf)-1,f)) return 0;
  buf[sizeof(buf)-1]= 0;
  for(i=0;buf[i]!=' ' && buf[i]!=0;i++) cmd[i]= buf[i];
  cmd[i]= 0;
  val[0]= 0;
  for(;buf[i]==' ';i++) ;
  if (!buf[i]) return 1;
  for(j=0;buf[i]!='\n' && buf[i]!=0;i++,j++) val[j]= buf[i];
  val[j]= 0;
  return 1;
}

#define OFFSET(x)  ((int)(size_t)(&(((bfile_t*)0)->x))) 

typedef struct { char *name; int offset; } str_field_t;

static str_field_t str_fields[]=
{
  {"C00COMPAT",    OFFSET(c_compiler_path) },
  {"C00COMOPT",    OFFSET(c_compiler_options) },
  {"C00COMOPTR00", OFFSET(c_compiler_optionsR) }, 
  {"C00COMOPTD00", OFFSET(c_compiler_optionsD) },

  {"C00LINOPT",    OFFSET(c_linker_options) },
  {"C00LINOPTR00", OFFSET(c_linker_optionsR) },
  {"C00LINOPTD00", OFFSET(c_linker_optionsD) },

  {"A00BUITOODIR", OFFSET(a_build_tool_dir) },
  {"A00JARFIL",    OFFSET(a_jarfile) },
  {"A00ZIPALI",    OFFSET(a_zipalign) },
  {"A00DX0",       OFFSET(a_dx) },
  {"A00AAP",       OFFSET(a_aapt) },
  {"A00JAV",       OFFSET(a_javac) },
  {"A00JARSIG",    OFFSET(a_jarsigner) },
  {"A00KEYTOO",    OFFSET(a_keytool) },
  {"A00PAC",       OFFSET(a_package) },
  {"A00APP",       OFFSET(a_appname) },
  {"A00ADB",       OFFSET(a_adb) },
  { NULL,          -1 }
};

static void save_string_fields(bfile_t *bf,FILE *out)
{
  int i;
  char **F;
  for(i=0;str_fields[i].name;i++)
  {
    F=  (char**) ((char*) bf + str_fields[i].offset);
    if (isempty(F[0])) continue;
    prt_cmd(out, str_fields[i].name, F[0]);
  }
}

static void save_list_field(bfile_t *bf, FILE *out,strlist_t *L,char *cmd)
{
  strlist_t *K;
  for(K=L->next;K!=L;K=K->next)
    prt_cmd(out, cmd, K->name);
}

static void save_c_sources(bfile_t *bf, FILE *out)
{
  c_source_t *s;
  c_dep_t *d;
  char buf[20];
  for(s=bf->c_sources.next;s!=&bf->c_sources;s=s->next)
  {
    sprintf(buf,"%d",s->output_no);
    prt_cmd(out, "C00NUM", buf);
    prt_cmd(out, "C00SOU", s->name);
    for(d=s->dhead.next;d!=&s->dhead;d=d->next)
     prt_cmd(out,"C00DEP", d->name);
  }
}


void bfile_save(bfile_t *bf)
{
  FILE *out;
  out= fopen(bf->fn,"wb");
  if (!out)
  {
    loprintf("can't open %s: %s\n", bf->fn, strerror(errno));
    return ;
  }  
  save_string_fields(bf, out);
  save_list_field(bf, out, &bf->c_include_paths, "INCLUD");
  save_list_field(bf, out, &bf->c_library_paths, "LIBRAR");
  save_list_field(bf, out, &bf->perms, "PERMIS");
  save_c_sources(bf, out);
  if (bf->scalign=='S') prt_cmd(out,"SCALIGN", "S");
                  else  prt_cmd(out,"SCALIGN", "L");
  fclose(out);
  loprintf("file saved %s\n",bf->fn);
}

static int load_string_field(bfile_t *bf, char *cmd, char *val)
{
  int i;
  char **F;
  for(i=0;str_fields[i].name;i++)
    if (!strcmp(str_fields[i].name, cmd))
    {
      F=  (char**) ((char*) bf + str_fields[i].offset);
      F[0]= strdup(val);
      return 1;
     }
  return 0;
}

static int lookup(const char **table, char *val)
{
  int i;
  for(i=0;table[i];i++) if (!strcmp(table[i],val)) return i;
  return -1;
}

bfile_t* bfile_open(char *path)
{
  bfile_t *bf;
  FILE *inp;
  int csno;
  c_source_t *cs;
  char cmd[2000], val[2000];

  bf= calloc(1,sizeof(*bf));
  if (!strcmp(path,"[Template]"))
    bf->fn= template_fn();
  else
    bf->fn= path_abs(path);
  bf->dispname= strdup(path);
  bf->dir= path_dir(bf->fn);

  strlist_init(&bf->perms);
  strlist_init(&bf->c_library_paths);
  strlist_init(&bf->c_include_paths);
  bf->c_sources.next= bf->c_sources.prev= &bf->c_sources;

  inp= fopen(bf->fn, "rb");
  if (!inp)
  {
    loprintf("can't open %s: %s\n", bf->fn, strerror(errno));
    goto done;
  }
  while(get_cmd(inp, cmd, val))
  {
    static const char *cmds[]=
    { "INCLUD","LIBRAR","PERMIS","SCALIGN","C00SOU","C00NUM","C00DEP", NULL };
    if (load_string_field(bf, cmd, val)) continue;
    switch(lookup(cmds,cmd))
    {
    case 0: bfile_add_c_include(bf, val); break;
    case 1: bfile_add_c_library(bf, val); break;
    case 2: bfile_add_perm(bf, val); break;
    case 3: bf->scalign= val[0]; break;
    case 4: cs= bfile_add_c_source(bf, val, csno); break;
    case 5: csno= atoi(val); break;
    case 6: if (cs) bfile_add_c_dependency(bf,cs,val); break;
    }
  }
  fclose(inp);
done:
  return bf;
}

void bfile_destroy(bfile_t *bf)
{
}

char* c_source_file(bfile_t *bf,char *name)
{
  return path_child(bf->dir, name);
}

char* c_object_file(bfile_t *bf, c_source_t *cs)
{
  char buf[200];
  sprintf(buf,"android/c_objects/out%d.o", cs->output_no);
  return path_child(bf->dir, buf);
}
