#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include "bfile.h"
#include "cmd.h"
#include "dep.h"
#include "main.h"
#include "path.h"

char *c_library_file(bfile_t *bf)
{
  return path_child(bf->dir, "android/lib/armeabi/libmain.so");
}

char *strcom2(char *a,char* b)
{
 char *R;
  R= malloc(strlen(a)+strlen(b)+1);
  sprintf(R,"%s%s",a,b);
  return R;
}

char *replace_suffix(char *str, char *sfx, char *nsfx)
{
  char *K;
  int L;
  if (strlen(sfx)>strlen(str)) return strdup(str);
  K= malloc(strlen(str) - strlen(sfx) + strlen(nsfx)+1);
  L= strlen(str) -strlen(sfx);
  memcpy(K, str, L);
  strcpy(K+L,nsfx);
  return K;
}

int check_field(char *f,char *name)
{
  if (isempty(f))
  {
    loprintf("please set %s\n",name);
    return 1;
  }
  return 0;
}

static int read_last_build(bfile_t *bf)
{
  char *fn;
  FILE *f;
  char g[200];
  fn= path_child(bf->dir, "android/last_build");
  f= fopen(fn, "rb");
  if (!f) return 'U';
  if (!fgets(g,sizeof(g),f)) { fclose(f); return 'U'; }
  fclose(f);
  return g[0];
}

static void write_last_build(bfile_t *bf,int dbg)
{
  char *fn;
  FILE *f;
  fn= path_child(bf->dir, "android/last_build");
  f= fopen(fn, "wb");
  if (!f) return ;
  switch(dbg)
  {
  case 'D':fprintf(f,"DEBUG\n"); break;
  case 'R':fprintf(f,"RELEASE\n"); break;
  default:fprintf(f,"UNKNOWN\n"); break;
  }
  fclose(f);
}

static int c_object_is_up_to_date(bfile_t *bf, c_source_t *cs)
{
  c_dep_t *cd;
  dep_reset();
  dep_add(c_source_file(bf,cs->name));
  for(cd=cs->dhead.next;cd!=&cs->dhead;cd=cd->next)
    dep_add(c_source_file(bf,cd->name));
  if (dep_end(c_object_file(bf, cs))) return 0;
  return 1;
}

static int check_compiler(bfile_t *bf)
{
  if (isempty(bf->c_compiler_path)) 
  {
    loprintf("you need to set the c compiler\n");
    return 1;
  }
  return 0;
}

static int compile_c_source(bfile_t *bf, c_source_t *cs,int force,int dbg)
{
  strlist_t *sn;
  char buf[2000];
  if (!force && c_object_is_up_to_date(bf, cs)) return 0;

  msg("Compiling %s",cs->name);
  if (check_compiler(bf)) return 1;
  cmd_reset();
  cmd_args(bf->c_compiler_path,NULL);
  cmd_parse(bf->c_compiler_options);
  cmd_parse(dbg=='D'? bf->c_compiler_optionsD:bf->c_compiler_optionsR);
  
  sprintf(buf,"-DGFX_ANDROID_ACTIVITY=\"%s/MyActivity\"",
       package_name_to_dir(bf->a_package));
  cmd_args(buf,"-DGFX_ANDROID", NULL);
  cmd_args("-o", c_object_file(bf, cs), "-c", c_source_file(bf,cs->name),NULL);

  for(sn=bf->c_include_paths.next;sn!=&bf->c_include_paths;sn=sn->next)
    cmd_args("-I", c_source_file(bf,sn->name),NULL);
  return cmd_run();
}

static int c_library_is_up_to_date(bfile_t *bf)
{
  c_source_t *cs;
  dep_reset();
  for(cs=bf->c_sources.next;cs!=&bf->c_sources;cs=cs->next)
    dep_add(c_object_file(bf,cs));
  if (dep_end(c_library_file(bf))) return 0;
  return 1;
}

static int link_c_objects(bfile_t *bf,int force,int dbg)
{
  strlist_t *sn;
  c_source_t *cs;

  if (!force && c_library_is_up_to_date(bf)) return 0;
  msg("Linking libmain.so");
  if (check_compiler(bf)) return 1;
  cmd_reset();
  cmd_args(bf->c_compiler_path,NULL);

  cmd_args("-o", c_library_file(bf), NULL);
  for(cs=bf->c_sources.next;cs!=&bf->c_sources;cs=cs->next)
    cmd_args(c_object_file(bf,cs),NULL);
  
  for(sn=bf->c_library_paths.next;sn!=&bf->c_library_paths;sn=sn->next)
    cmd_args("-L", sn->name,NULL);

  cmd_parse(bf->c_linker_options);
  cmd_parse(dbg=='D'? bf->c_linker_optionsD:bf->c_linker_optionsR);
  return cmd_run();
}

int compile_c_sources(bfile_t *bf,int dbg)
{
  int R;
  int force;
  c_source_t *cs;

  if (read_last_build(bf)!=dbg) force= 1;
                           else force= 0;

  R= 1;
  for(cs=bf->c_sources.next;cs!=&bf->c_sources;cs=cs->next)
   if (compile_c_source(bf, cs, force, dbg)) 
       goto done; 

  R= link_c_objects(bf,force,dbg);
  
done:
  if (R) { write_last_build(bf,'U'); msg("Failed"); }
    else { write_last_build(bf,dbg); msg("All done"); }
  return R;
}

char *android_tool(bfile_t *bf, char *ref,char *name)
{
  if (!isempty(ref)) return ref;
  return path_child(bf->a_build_tool_dir, name);
}

char *bfile_resources_dir(bfile_t *bf)
{
  return path_child(bf->dir, "android/resources");
}

char *bfile_java_source_root(bfile_t *bf)
{
  return path_child(bf->dir, "android/java_sources");
}

char *bfile_manifest_path(bfile_t *bf)
{
  return path_child(bf->dir, "android/AndroidManifest.xml");
}

char *bfile_rjava_path(bfile_t *bf)
{
  char *dir;
  char *pakdir;
  pakdir= package_name_to_dir(bf->a_package);
  dir= path_child(bfile_java_source_root(bf),pakdir);
  return path_child(dir, "R.java");
}

int make_resources(bfile_t *bf)
{
  char **resfiles;
  int i;
  resfiles= find_files(bfile_resources_dir(bf), NULL);
  dep_reset();
  for(i=0;resfiles[i];i++)
    dep_add(resfiles[i]);
  dep_add(bfile_manifest_path(bf));
  if (!dep_end(bfile_rjava_path(bf))) return 0;
  msg("Compiling resources");

  if (check_field(bf->a_package,"package")) return 1;
  if (check_field(bf->a_jarfile,"jarfile")) return 1;
  if (check_field(bf->a_build_tool_dir, "build_tool_dir")) return 1;

  cmd_reset();
  cmd_args( android_tool(bf, bf->a_aapt, "aapt"), "package", "-f", "-m",
            "-S", bfile_resources_dir(bf), 
            "-J", bfile_java_source_root(bf),
            "-M", bfile_manifest_path(bf),
            "-I", bf->a_jarfile, NULL);
  return cmd_run();
}



char *bfile_java_object(bfile_t *bf, char *src_path)
{
  return path_child(java_object_dir(bf),
           replace_suffix(path_file(src_path),".java",".class"));
}

static char *bfile_java_out(bfile_t *bf)
{
  return path_child(bf->dir,"android/java_objects");
}

static char *bfile_dex(bfile_t *bf)
{
 return path_child(bf->dir,"android/bin/classes.dex");
}

static int compile_java_source(bfile_t *bf,char *src_path)
{
  dep_reset();
  dep_add(src_path);
  if (!dep_end(bfile_java_object(bf,src_path))) return 0;
  msg("Compiling %s", path_file(src_path));
  cmd_reset();
  cmd_args( system_tool(bf, bf->a_javac, "javac"),
            "-d", bfile_java_out(bf),
            "-classpath", bf->a_jarfile,
            "-sourcepath", bfile_java_source_root(bf),
            src_path, bfile_rjava_path(bf), NULL);
  return cmd_run();
}

int compile_java_sources(bfile_t *bf)
{
  char **src;
  int i;
  src= find_files(bfile_java_source_root(bf), ".java");
  for(i=0;src[i];i++)
    if (compile_java_source(bf, src[i]))
       return 1;
  return 0;
}


int build_dex(bfile_t *bf)
{
  char buf[2000];
  char **src;
  int i;
  src= find_files(bfile_java_source_root(bf), ".java");
  dep_reset();
  for(i=0;src[i];i++)
    dep_add(bfile_java_object(bf,src[i]));
  if (!dep_end(bfile_dex(bf))) return 0;
  msg("Building classes.dex");
  cmd_reset();
  sprintf(buf,"--output=%s",bfile_dex(bf));
  cmd_args( android_tool(bf,bf->a_dx, "dx"), 
            "--dex", buf, bfile_java_out(bf), NULL);
  return cmd_run();
}

int build_apk(bfile_t *bf)
{
  char *u_apk, *s_apk, *z_apk;

  u_apk= strcom2(bf->a_appname, ".unsigned.apk");
  s_apk= strcom2(bf->a_appname, ".signed.apk");
  z_apk= strcom2(bf->a_appname, ".apk");

  if (chdir(path_child(bf->dir,"android")))
  {
    int e= errno;
    loprintf("chdir(%s): %s\n", path_child(bf->dir,"android"),strerror(e));
    return 1;
  }
  dep_reset();
  dep_add("bin/classes.dex");
  dep_add("lib/armeabi/libmain.so");
  if (!dep_end(u_apk)) goto next1;

  msg("Building APK: %s", u_apk);
  cmd_reset();
  cmd_args(android_tool(bf, bf->a_aapt, "aapt"), "package",
           "-f", "-M", "AndroidManifest.xml",
           "-S", "resources", 
           "-I", bf->a_jarfile,
           "-F", u_apk, "bin/", NULL);
  if (cmd_run()) return 1;

next1:

  dep_reset(); dep_add(u_apk); if (!dep_end(s_apk)) goto next2;
  msg("Sign APK: %s", u_apk);
  
  cmd_reset();
  cmd_args(system_tool(bf,bf->a_jarsigner, "jarsigner"),
           "-keystore", "test.keystore",
           "-storepass", "haxorz", "-keypass", "haxorz",
           "-signedjar", s_apk, u_apk, "helljniKey", NULL);
  if (cmd_run()) return 1;

next2:
  dep_reset(); dep_add(s_apk); if (!dep_end(z_apk)) goto next3;
  msg("Zipalign %s", s_apk);
  cmd_reset();
  cmd_args( android_tool(bf, bf->a_zipalign, "zipalign"),
             "-f" , "4", s_apk, z_apk, NULL);
  if (cmd_run()) return 1;
next3:
  loprintf("APK %s ready\n", z_apk);
  return 0;
}
